import ast
import shutil
from contextlib import nullcontext
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Set, Union
from urllib.parse import urlparse

import httpx
from autoregistry import Registry
from rich.console import Console


class NonMatchingURI(Exception):
    pass


uri_processors = Registry()


@dataclass
class GroupConfig:
    """Schema and store of a group defined in ``pyproject.toml``.

    Don't put any methods in here, they go in ``Group``.
    This class is primarily for namespacing and validation.
    """

    name: str
    optional: bool = False
    dependencies: Dict[str, str] = field(default_factory=dict)


class Group:
    """Represents a group defined in ``pyproject.toml``."""

    def __init__(self, *args, **kwargs):
        from belay.project import find_dependencies_folder

        self.config = GroupConfig(*args, **kwargs)

        self.folder = find_dependencies_folder() / self.config.name

        if self.config.optional:
            raise NotImplementedError("Optional groups not implemented yet.")

    def __eq__(self, other):
        if not isinstance(other, Group):
            return False
        return self.config.__dict__ == other.config.__dict__

    def __repr__(self):
        kws = [f"{key}={value!r}" for key, value in self.config.__dict__.items()]
        return f"{type(self).__name__}({', '.join(kws)})"

    def clean(self):
        """Delete any dependency module not specified in ``self.config.dependencies``."""
        dependencies = set(self.config.dependencies)
        existing_deps = []

        if not self.folder.exists():
            return

        existing_deps.extend(self.folder.glob("*"))

        for existing_dep in existing_deps:
            if existing_dep.stem in dependencies:
                continue
            existing_dep.unlink()

    def copy_to(self, dst):
        """Copy Dependencies folder to destination directory."""
        if self.folder.exists():
            shutil.copytree(self.folder, dst, dirs_exist_ok=True)

    def download(
        self,
        packages: Optional[List[str]] = None,
        console: Optional[Console] = None,
    ):
        """Download dependencies.

        Parameters
        ----------
        packages: Optional[List[str]]
            Only download these package.
        console: Optional[Console]
            Print progress out to console.
        """
        if packages is None:
            # Update all packages
            packages = list(self.config.dependencies.keys())

        if not packages:
            return

        if console:
            cm = console.status("[bold green]Updating Dependencies")
        else:
            cm = nullcontext()

        def log(*args, **kwargs):
            if console:
                console.log(*args, **kwargs)

        with cm:
            for package_name in packages:
                dep_src = self.config.dependencies[package_name]
                if isinstance(dep_src, str):
                    dep_src = {"path": dep_src}
                elif not isinstance(dep_src, dict):
                    raise ValueError(f"Invalid value for key {package_name}.")

                log(f"{package_name}: Updating...")

                uri = _process_uri(dep_src["path"])
                ext = Path(uri).suffix

                # Single file
                dst = self.folder / (package_name + ext)
                dst.parent.mkdir(parents=True, exist_ok=True)

                new_code = _get_text(uri)

                if ext == ".py":
                    ast.parse(new_code)  # Check for valid python code

                try:
                    old_code = dst.read_text()
                except FileNotFoundError:
                    old_code = ""

                if new_code == old_code:
                    log(f"{package_name}: No changes detected.")
                else:
                    log(f"[bold green]{package_name}: Updated.")
                    dst.write_text(new_code)


def _strip_www(uri: str):
    if uri.startswith("www."):
        uri = uri[4:]
    return uri


@uri_processors
def _process_uri_github(uri: str):
    """Transforms github-like uri into githubusercontent."""
    uri = str(uri)
    parsed = urlparse(uri)
    netloc = _strip_www(parsed.netloc)
    if netloc == "github.com":
        # Transform to raw.githubusercontent
        _, user, project, mode, branch, *path = parsed.path.split("/")
        return f"https://raw.githubusercontent.com/{user}/{project}/{branch}/{'/'.join(path)}"
    elif netloc == "raw.githubusercontent.com":
        return f"https://raw.githubusercontent.com{parsed.path}"
    else:
        # TODO: Try and be a little helpful if uri contains github.com
        raise NonMatchingURI


def _process_uri(uri: str):
    for processor in uri_processors.values():
        try:
            return processor(uri)
        except NonMatchingURI:
            pass

    # Unmodified URI
    return uri


def _get_text(uri: Union[str, Path]):
    uri = str(uri)
    if uri.startswith(("https://", "http://")):
        res = httpx.get(uri)
        res.raise_for_status()
        return res.text
    else:
        # Assume local file
        return Path(uri).read_text()



import os
import shutil
import subprocess  # nosec
import sys
from typing import List

import typer

from belay.cli.clean import clean
from belay.cli.exec import exec
from belay.cli.identify import identify
from belay.cli.info import info
from belay.cli.install import install
from belay.cli.new import new
from belay.cli.run import run
from belay.cli.sync import sync
from belay.cli.update import update
from belay.project import load_groups

app = typer.Typer()
app.command()(clean)
app.command()(exec)
app.command()(identify)
app.command()(info)
app.command()(install)
app.command()(new)
app.command()(run)
app.command()(sync)
app.command()(update)


def run_exec(command: List[str]):
    """Pseudo-virtual-environment."""
    groups = load_groups()
    virtual_env = os.environ.copy()
    # Add all dependency groups to the micropython path.
    virtual_env["MICROPYPATH"] = os.pathsep.join(str(g.folder) for g in groups)
    return subprocess.run(  # nosec
        command,
        env=virtual_env,
        check=True,
    ).returncode


def _get(indexable, index, default=None):
    try:
        return indexable[index]
    except IndexError:
        return default


def run_app(*args, **kwargs):
    """Add CLI hacks that are not Typer-friendly here."""
    command = _get(sys.argv, 1)

    try:
        exec_path = shutil.which(sys.argv[2])
    except IndexError:
        exec_path = None

    if command == "run" and exec_path:
        # See docstring in belay.cli.run.run
        return run_exec(sys.argv[2:])

    # Common-case; use Typer functionality.
    app(*args, **kwargs)



import importlib.resources as pkg_resources
import secrets
import string
from functools import lru_cache, partial, wraps
from typing import List

from serial.tools import list_ports

from . import snippets

_python_identifier_chars = (
    string.ascii_uppercase + string.ascii_lowercase + string.digits
)


def wraps_partial(f, *args, **kwargs):
    """Wrap and partial of a function."""
    return wraps(f)(partial(f, *args, **kwargs))


def random_python_identifier(n=16):
    return "_" + "".join(secrets.choice(_python_identifier_chars) for _ in range(n))


@lru_cache
def read_snippet(name):
    return pkg_resources.read_text(snippets, f"{name}.py")


def list_devices() -> List[str]:
    """Lists available device ports.

    For example::

        ['/dev/cu.usbmodem1143401', '/dev/cu.usbmodem113101']

    Returns
    -------
    list[str]
        Available devices identifiers.
    """
    return [port.device for port in list_ports.comports()]



from functools import lru_cache
from pathlib import Path
from typing import List, Union

import tomli

from belay.exceptions import ConfigError
from belay.packagemanager import Group


@lru_cache
def find_pyproject() -> Path:
    path = Path("pyproject.toml").absolute()

    for parent in path.parents:
        candidate = parent / path.name
        if candidate.exists():
            return candidate
    raise FileNotFoundError(
        f'Cannot find a pyproject.toml in the current directory "{Path().absolute()}" or any parent directory.'
    )


@lru_cache
def find_project_folder() -> Path:
    return find_pyproject().parent


@lru_cache
def find_belay_folder() -> Path:
    return find_project_folder() / ".belay"


@lru_cache
def find_dependencies_folder() -> Path:
    return find_belay_folder() / "dependencies"


@lru_cache
def load_toml(path: Union[str, Path]) -> dict:
    path = Path(path)
    with path.open("rb") as f:
        toml = tomli.load(f)

    try:
        toml = toml["tool"]["belay"]
    except KeyError:
        return {}

    return toml


@lru_cache
def load_pyproject() -> dict:
    """Load the pyproject TOML file."""
    pyproject_path = find_pyproject()
    return load_toml(pyproject_path)


@lru_cache
def load_groups() -> List[Group]:
    config = load_pyproject()
    groups_definitions = config.get("group", {})
    if "main" in groups_definitions:
        raise ConfigError(
            'Specify "main" group dependencies under "tool.belay.dependencies", '
            'not "tool.belay.group.main.dependencies"'
        )
    if "dependencies" in config:
        groups_definitions["main"] = {"dependencies": config["dependencies"]}

    groups = [
        Group(name, **definition) for name, definition in groups_definitions.items()
    ]
    groups.sort(key=lambda x: x.config.name)

    return groups
